# Copyright 2017 Cloudbase Solutions SRL
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""This module contains the L2 Agent needed for HNV."""

import platform
import sys

from neutron.common import config as common_config
from neutron.conf.agent import common as neutron_config
from oslo_log import log as logging

from networking_hyperv.common.i18n import _LI    # noqa
from networking_hyperv.neutron.agent import layer2 as hyperv_base
from networking_hyperv.neutron import config
from networking_hyperv.neutron import constants as h_const
from networking_hyperv.neutron import neutron_client

LOG = logging.getLogger(__name__)
CONF = config.CONF


class HNVAgent(hyperv_base.Layer2Agent):

    _AGENT_BINARY = "neutron-hnv-agent"
    _AGENT_TYPE = h_const.AGENT_TYPE_HNV

    def __init__(self):
        super(HNVAgent, self).__init__()
        # Handle updates from service
        self._agent_id = 'hnv_%s' % platform.node()
        self._neutron_client = neutron_client.NeutronAPIClient()

    def _get_agent_configurations(self):
        return {
            'logical_network': CONF.HNV.logical_network,
            'vswitch_mappings': self._physical_network_mappings,
            'devices': 1,
            'l2_population': False,
            'tunnel_types': [],
            'bridge_mappings': {},
            'enable_distributed_routing': False,
        }

    def _provision_network(self, port_id, net_uuid, network_type,
                           physical_network, segmentation_id):
        """Provision the network with the received information."""
        LOG.info("Provisioning network %s", net_uuid)

        vswitch_name = self._get_vswitch_name(network_type, physical_network)
        vswitch_map = {
            'network_type': network_type,
            'vswitch_name': vswitch_name,
            'ports': [],
            'vlan_id': segmentation_id}
        self._network_vswitch_map[net_uuid] = vswitch_map

    def _port_bound(self, port_id, network_id, network_type, physical_network,
                    segmentation_id, port_security_enabled, set_port_sriov):
        """Bind the port to the recived network."""
        super(HNVAgent, self)._port_bound(port_id, network_id, network_type,
                                          physical_network, segmentation_id,
                                          port_security_enabled,
                                          set_port_sriov)
        LOG.debug("Getting the profile id for the current port.")
        profile_id = self._neutron_client.get_port_profile_id(port_id)

        LOG.debug("Trying to set port profile id %r for the current port %r.",
                  profile_id, port_id)
        self._utils.set_vswitch_port_profile_id(
            switch_port_name=port_id,
            profile_id=profile_id,
            profile_data=h_const.PROFILE_DATA,
            profile_name=h_const.PROFILE_NAME,
            net_cfg_instance_id=h_const.NET_CFG_INSTANCE_ID,
            cdn_label_id=h_const.CDN_LABEL_ID,
            cdn_label_string=h_const.CDN_LABEL_STRING,
            vendor_id=h_const.VENDOR_ID,
            vendor_name=h_const.VENDOR_NAME)


def main():
    """The entry point for the HNV Agent."""
    neutron_config.register_agent_state_opts_helper(CONF)
    common_config.init(sys.argv[1:])
    neutron_config.setup_logging()

    hnv_agent = HNVAgent()

    # Start everything.
    LOG.info("Agent initialized successfully, now running... ")
    hnv_agent.daemon_loop()



# Copyright 2017 Cloudbase Solutions Srl
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from neutron.agent.l2.extensions import qos
from neutron_lib.services.qos import constants as qos_consts
from os_win.utils.network import networkutils
from oslo_log import log as logging

from networking_hyperv.common.i18n import _LI, _LW  # noqa

LOG = logging.getLogger(__name__)


class QosHyperVAgentDriver(qos.QosAgentDriver):

    _SUPPORTED_QOS_RULES = [qos_consts.RULE_TYPE_BANDWIDTH_LIMIT,
                            qos_consts.RULE_TYPE_MINIMUM_BANDWIDTH]

    def initialize(self):
        self._utils = networkutils.NetworkUtils()

    def create(self, port, qos_policy):
        """Apply QoS rules on port for the first time.

        :param port: port object.
        :param qos_policy: the QoS policy to be applied on port.
        """
        LOG.info("Setting QoS policy %(qos_policy)s on port %(port)s",
                 dict(qos_policy=qos_policy, port=port))

        policy_data = self._get_policy_values(qos_policy)
        self._utils.set_port_qos_rule(port["port_id"], policy_data)

    def update(self, port, qos_policy):
        """Apply QoS rules on port.

        :param port: port object.
        :param qos_policy: the QoS policy to be applied on port.
        """
        LOG.info("Updating QoS policy %(qos_policy)s on port %(port)s",
                 dict(qos_policy=qos_policy, port=port))

        policy_data = self._get_policy_values(qos_policy)
        self._utils.set_port_qos_rule(port["port_id"], policy_data)

    def delete(self, port, qos_policy=None):
        """Remove QoS rules from port.

        :param port: port object.
        :param qos_policy: the QoS policy to be removed from port.
        """
        LOG.info("Deleting QoS policy %(qos_policy)s on port %(port)s",
                 dict(qos_policy=qos_policy, port=port))

        self._utils.remove_port_qos_rule(port["port_id"])

    def _get_policy_values(self, qos_policy):
        result = {}
        for qos_rule in qos_policy.rules:
            if qos_rule.rule_type not in self._SUPPORTED_QOS_RULES:
                LOG.warning("Unsupported QoS rule: %(qos_rule)s",
                            dict(qos_rule=qos_rule))
                continue
            result['min_kbps'] = getattr(qos_rule, 'min_kbps',
                                         result.get('min_kbps'))
            result['max_kbps'] = getattr(qos_rule, 'max_kbps',
                                         result.get('max_kbps'))
            result['max_burst_kbps'] = getattr(qos_rule, 'max_burst_kbps',
                                               result.get('max_burst_kbps'))

        return result



# Copyright 2015 Cloudbase Solutions Srl
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from keystoneauth1 import loading as ks_loading
from oslo_config import cfg

from networking_hyperv.common.i18n import _

CONF = cfg.CONF


HYPERV_AGENT_GROUP_NAME = 'AGENT'

HYPERV_AGENT_GROUP = cfg.OptGroup(
    HYPERV_AGENT_GROUP_NAME,
    title='Hyper-V Neutron Agent Options',
    help=('Configuration options for the neutron-hyperv-agent (L2 agent).')
)

HYPERV_AGENT_OPTS = [
    cfg.ListOpt(
        'physical_network_vswitch_mappings',
        default=[],
        help=_('List of <physical_network>:<vswitch> '
               'where the physical networks can be expressed with '
               'wildcards, e.g.: ."*:external"')),
    cfg.StrOpt(
        'local_network_vswitch',
        help=_('Private vswitch name used for local networks')),
    cfg.IntOpt('polling_interval', default=2, min=1,
               help=_("The number of seconds the agent will wait between "
                      "polling for local device changes.")),
    cfg.IntOpt('worker_count', default=10, min=1,
               help=_("The number of worker threads allowed to run in "
                      "parallel to process port binding.")),
    cfg.IntOpt('worker_retry', default=3, min=0,
               help=_("The number of times worker process will retry "
                      "port binding.")),
    cfg.BoolOpt('enable_metrics_collection',
                default=False,
                help=_('Enables metrics collections for switch ports by using '
                       'Hyper-V\'s metric APIs. Collected data can by '
                       'retrieved by other apps and services, e.g.: '
                       'Ceilometer. Requires Hyper-V / Windows Server 2012 '
                       'and above')),
    cfg.IntOpt('metrics_max_retries',
               default=100, min=0,
               help=_('Specifies the maximum number of retries to enable '
                      'Hyper-V\'s port metrics collection. The agent will try '
                      'to enable the feature once every polling_interval '
                      'period for at most metrics_max_retries or until it '
                      'succeedes.')),
    cfg.IPOpt('neutron_metadata_address',
              default='169.254.169.254',
              help=_('Specifies the address which will serve the metadata for'
                      ' the instance.')),
    cfg.BoolOpt('enable_qos_extension',
                default=False,
                help=_('Enables the QoS extension.')),
]


NVGRE_GROUP_NAME = 'NVGRE'

NVGRE_GROUP = cfg.OptGroup(
    NVGRE_GROUP_NAME,
    title='Hyper-V NVGRE Options',
    help=('Configuration options for NVGRE.')
)

NVGRE_OPTS = [
    cfg.BoolOpt('enable_support',
                default=False,
                help=_('Enables Hyper-V NVGRE. '
                       'Requires Windows Server 2012 or above.')),
    cfg.IntOpt('provider_vlan_id',
               default=0, min=0, max=4096,
               help=_('Specifies the VLAN ID of the physical network, required'
                      ' for setting the NVGRE Provider Address.')),
    cfg.IPOpt('provider_tunnel_ip',
              default=None,
              help=_('Specifies the tunnel IP which will be used and '
                     'reported by this host for NVGRE networks.')),
]


NEUTRON_GROUP_NAME = 'neutron'

NEUTRON_GROUP = cfg.OptGroup(
    NEUTRON_GROUP_NAME,
    title='Neutron Options',
    help=('Configuration options for neutron (network connectivity as a '
          'service).')
)

NEUTRON_OPTS = [
    cfg.StrOpt('url',
               default='http://127.0.0.1:9696',
               help='URL for connecting to neutron'),
    cfg.IntOpt('url_timeout',
               default=30, min=1,
               help='timeout value for connecting to neutron in seconds'),
    cfg.StrOpt('admin_username',
               help='username for connecting to neutron in admin context'),
    cfg.StrOpt('admin_password',
               help='password for connecting to neutron in admin context',
               secret=True),
    cfg.StrOpt('admin_tenant_name',
               help='tenant name for connecting to neutron in admin context'),
    cfg.StrOpt('admin_auth_url',
               default='http://localhost:5000/v2.0',
               help='auth url for connecting to neutron in admin context'),
    cfg.StrOpt('auth_strategy',
               default='keystone',
               help='auth strategy for connecting to neutron in admin context')
]


HNV_GROUP_NAME = 'HNV'

HNV_GROUP = cfg.OptGroup(
    HNV_GROUP_NAME,
    title='HNV Options',
    help='Configuration options for the Windows Network Controller.'
)

HNV_OPTS = [
    cfg.StrOpt(
        "logical_network", default=None,
        help=("Logical network to use as a medium for tenant network "
              "traffic.")),
]

ALL_OPTS = [
    (HYPERV_AGENT_GROUP, HYPERV_AGENT_OPTS),
    (NVGRE_GROUP, NVGRE_OPTS),
    (NEUTRON_GROUP, NEUTRON_OPTS),
    (HNV_GROUP, HNV_OPTS)
]


def register_opts():
    for group, opts in ALL_OPTS:
        CONF.register_group(group)
        CONF.register_opts(opts, group=group)

    ks_loading.register_session_conf_options(CONF, NEUTRON_GROUP)
    ks_loading.register_auth_conf_options(CONF, NEUTRON_GROUP)


def list_opts():
    return ALL_OPTS


register_opts()



# Copyright 2013 Cloudbase Solutions SRL
# Copyright 2013 Pedro Navarro Perez
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import inspect
from os_win import exceptions as os_win_exc
from oslo_concurrency import lockutils
from oslo_log import log as logging
from oslo_utils import reflection

LOG = logging.getLogger(__name__)


def get_port_synchronized_decorator(lock_prefix):
    synchronized = lockutils.synchronized_with_prefix(lock_prefix)

    def _port_synchronized(f):
        # This decorator synchronizes operations targeting the same port.
        # The decorated method is expected to accept the port_id argument.
        def wrapper(*args, **kwargs):
            call_args = inspect.getcallargs(f, *args, **kwargs)
            port_id = (call_args.get('port_id') or
                       call_args.get('port', {}).get('id'))
            lock_name = lock_prefix + ('port-lock-%s' % port_id)

            @synchronized(lock_name)
            def inner():
                return f(*args, **kwargs)
            return inner()
        return wrapper
    return _port_synchronized


def ignore_missing_ports(f):
    def wrapper(*args, **kwargs):
        # The decorated method is expected to accept a port as argument.
        call_args = inspect.getcallargs(f, *args, **kwargs)
        port_id = (call_args.get('port_id') or
                   call_args.get('port', {}).get('id'))

        try:
            return f(*args, **kwargs)
        except os_win_exc.NotFound:
            func_name = reflection.get_callable_name(f)
            LOG.warning("Could not find port '%(port_id)s' while executing "
                        "'%(func_name)s'. It may have been removed.",
                        dict(port_id=port_id,
                             func_name=func_name))
    return wrapper



# Copyright 2017 Cloudbase Solutions Srl
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
import sys

import mock
from neutron.agent import rpc as agent_rpc
from neutron import wsgi
from neutron_lib.agent import topics
from oslo_config import cfg
import webob

from networking_hyperv.neutron.agent import base as base_agent
from networking_hyperv.neutron.agent import hnv_metadata_agent
from networking_hyperv.tests import base as test_base

CONF = cfg.CONF


class TestMetadataProxyHandler(test_base.BaseTestCase):

    _autospec_classes = [
        hnv_metadata_agent.neutron_client.NeutronAPIClient,
    ]

    def setUp(self):
        super(TestMetadataProxyHandler, self).setUp()
        hnv_metadata_agent.register_config_opts()
        self._proxy = hnv_metadata_agent._MetadataProxyHandler()
        self._neutron_client = self._proxy._neutron_client

    @mock.patch.object(hnv_metadata_agent._MetadataProxyHandler,
                       "_proxy_request")
    def test_call(self, mock_proxy_request):
        mock_proxy_request.side_effect = [mock.sentinel.response,
                                          ValueError("_proxy_request_error")]

        self.assertEqual(mock.sentinel.response,
                         self._proxy(mock.sentinel.request))
        mock_proxy_request.assert_called_once_with(mock.sentinel.request)

        self.assertIsInstance(self._proxy(mock.sentinel.request),
                              webob.exc.HTTPInternalServerError)

    def test_get_port_profile_id(self):
        url = "http://169.254.169.254/"
        port_profile_id = "9d0bab3e-1abf-11e7-a7ef-5cc5d4a321db"
        request = mock.Mock(path=url + port_profile_id)
        request_invalid = mock.Mock(path=url)

        self.assertEqual(port_profile_id,
                         self._proxy._get_port_profile_id(request))
        self.assertIsNone(self._proxy._get_port_profile_id(request_invalid))

    def test_get_instance_id(self):
        self._neutron_client.get_network_ports.return_value = [
            {},
            {"binding:vif_details": {"port_profile_id": None}},
            {"binding:vif_details": {
                "port_profile_id": mock.sentinel.port_profile_id},
             "tenant_id": mock.sentinel.tenant_id,
             "device_id": mock.sentinel.instance_id},
        ]
        self.assertEqual(
            (mock.sentinel.tenant_id, mock.sentinel.instance_id),
            self._proxy._get_instance_id(mock.sentinel.port_profile_id))

        self._neutron_client.get_network_ports.return_value = []
        self.assertEqual(
            (None, None),
            self._proxy._get_instance_id(mock.sentinel.port_profile_id))

    def test_sign_instance_id(self):
        self.config(metadata_proxy_shared_secret="secret")
        self.assertEqual(
            "0329a06b62cd16b33eb6792be8c60b158d89a2ee3a876fce9a881ebb488c0914",
            self._proxy._sign_instance_id("test")
        )

    @mock.patch.object(hnv_metadata_agent._MetadataProxyHandler,
                       "_sign_instance_id")
    @mock.patch.object(hnv_metadata_agent._MetadataProxyHandler,
                       "_get_instance_id")
    def test_get_headers(self, mock_get_instance_id, mock_sign_instance_id):
        mock_get_instance_id.side_effect = [
            (mock.sentinel.tenant_id, mock.sentinel.instance_id),
            (None, None),
        ]
        expected_headers = {
            'X-Instance-ID': mock.sentinel.instance_id,
            'X-Tenant-ID': mock.sentinel.tenant_id,
            'X-Instance-ID-Signature': mock_sign_instance_id.return_value,
        }

        self.assertEqual(
            expected_headers,
            self._proxy._get_headers(mock.sentinel.port))
        mock_get_instance_id.assert_called_once_with(mock.sentinel.port)
        self.assertIsNone(self._proxy._get_headers(mock.sentinel.port))

    @mock.patch("httplib2.Http")
    @mock.patch.object(hnv_metadata_agent._MetadataProxyHandler,
                       "_get_headers")
    def _test_proxy_request(self, mock_get_headers, mock_http,
                            valid_path=True, valid_profile_id=True,
                            response_code=200, method='GET'):
        nova_url = '%s:%s' % (CONF.nova_metadata_host,
                              CONF.nova_metadata_port)
        path = "/9d0bab3e-1abf-11e7-a7ef-5cc5d4a321db" if valid_path else "/"
        headers = {"X-Not-Empty": True} if valid_profile_id else {}
        mock_get_headers.return_value = headers

        http_response = mock.MagicMock(status=response_code)
        http_response.__getitem__.return_value = "text/plain"
        http_request = mock_http.return_value
        http_request.request.return_value = (http_response,
                                             mock.sentinel.content)

        mock_resonse = mock.Mock(content_type=None, body=None)
        mock_request = mock.Mock(path=path, path_info=path, query_string='',
                                 headers={}, method=method,
                                 body=mock.sentinel.body)
        mock_request.response = mock_resonse

        response = self._proxy._proxy_request(mock_request)

        if not (valid_path and valid_profile_id):
            http_request.add_certificate.assert_not_called()
            http_request.request.assert_not_called()
            return response

        if CONF.nova_client_cert and CONF.nova_client_priv_key:
            http_request.add_certificate.assert_called_once_with(
                key=CONF.nova_client_priv_key,
                cert=CONF.nova_client_cert,
                domain=nova_url)

        http_request.request.assert_called_once_with(
            "http://127.0.0.1:8775/", method=method, headers=headers,
            body=mock.sentinel.body)

        return response

    def test_proxy_request_200(self):
        self.config(nova_client_cert=mock.sentinel.nova_client_cert,
                    nova_client_priv_key=mock.sentinel.priv_key)
        response = self._test_proxy_request()
        self.assertEqual("text/plain", response.content_type)
        self.assertEqual(mock.sentinel.content, response.body)

    def test_proxy_request_400(self):
        self.assertIsInstance(
            self._test_proxy_request(response_code=400),
            webob.exc.HTTPBadRequest)

    def test_proxy_request_403(self):
        self.assertIsInstance(
            self._test_proxy_request(response_code=403),
            webob.exc.HTTPForbidden)

    def test_proxy_request_409(self):
        self.assertIsInstance(
            self._test_proxy_request(response_code=409),
            webob.exc.HTTPConflict)

    def test_proxy_request_404(self):
        self.assertIsInstance(
            self._test_proxy_request(valid_path=False),
            webob.exc.HTTPNotFound)
        self.assertIsInstance(
            self._test_proxy_request(valid_profile_id=False),
            webob.exc.HTTPNotFound)
        self.assertIsInstance(
            self._test_proxy_request(response_code=404),
            webob.exc.HTTPNotFound)

    def test_proxy_request_500(self):
        self.assertIsInstance(
            self._test_proxy_request(response_code=500),
            webob.exc.HTTPInternalServerError)

    def test_proxy_request_other_code(self):
        self.assertIsInstance(
            self._test_proxy_request(response_code=527),
            webob.exc.HTTPInternalServerError)

    def test_proxy_request_post(self):
        response = self._test_proxy_request(method='POST')
        self.assertEqual("text/plain", response.content_type)
        self.assertEqual(mock.sentinel.content, response.body)


class TestMetadataProxy(test_base.HyperVBaseTestCase):

    @mock.patch.object(hnv_metadata_agent.MetadataProxy, "_setup_rpc")
    @mock.patch.object(base_agent.BaseAgent, "_set_agent_state")
    def _get_agent(self, mock_set_agent_state, mock_setup_rpc):
        return hnv_metadata_agent.MetadataProxy()

    def setUp(self):
        super(TestMetadataProxy, self).setUp()
        hnv_metadata_agent.register_config_opts()
        self._agent = self._get_agent()

    @mock.patch('oslo_service.loopingcall.FixedIntervalLoopingCall')
    @mock.patch.object(agent_rpc, 'PluginReportStateAPI')
    def test_setup_rpc(self, mock_plugin_report_state_api,
                       mock_looping_call):
        report_interval = 10
        self.config(report_interval=report_interval, group="AGENT")

        self._agent._setup_rpc()

        mock_plugin_report_state_api.assert_called_once_with(topics.REPORTS)
        mock_looping_call.assert_called_once_with(self._agent._report_state)
        mock_heartbeat = mock_looping_call.return_value
        mock_heartbeat.start.assert_called_once_with(interval=report_interval)

    def test_get_agent_configurations(self):
        fake_ip = '10.10.10.10'
        fake_port = 9999
        self.config(nova_metadata_host=fake_ip,
                    nova_metadata_port=fake_port)

        configuration = self._agent._get_agent_configurations()

        self.assertEqual(fake_ip, configuration["nova_metadata_host"])
        self.assertEqual(fake_port, configuration["nova_metadata_port"])
        self.assertEqual(CONF.AGENT.log_agent_heartbeats,
                         configuration["log_agent_heartbeats"])

    @mock.patch.object(hnv_metadata_agent, "_MetadataProxyHandler")
    @mock.patch.object(wsgi, "Server")
    def test_work(self, mock_server, mock_proxy_handler):
        self._agent._work()

        mock_server.assert_called_once_with(
            name=self._agent._AGENT_BINARY,
            num_threads=CONF.AGENT.worker_count)
        server = mock_server.return_value
        server.start.assert_called_once_with(
            application=mock_proxy_handler.return_value,
            port=CONF.bind_port,
            host=CONF.bind_host)
        server.wait.assert_called_once_with()

    @mock.patch.object(hnv_metadata_agent.MetadataProxy, "_work")
    @mock.patch.object(hnv_metadata_agent.MetadataProxy, "_prologue")
    def test_run(self, mock_prologue, mock_work):
        mock_work.side_effect = ValueError
        self._agent.run()

        mock_prologue.assert_called_once_with()
        mock_work.assert_called_once_with()


class TestMain(test_base.BaseTestCase):

    @mock.patch.object(hnv_metadata_agent, 'MetadataProxy')
    @mock.patch.object(hnv_metadata_agent, 'common_config')
    @mock.patch.object(hnv_metadata_agent, 'meta_config')
    @mock.patch.object(hnv_metadata_agent, 'neutron_config')
    def test_main(self, mock_config, mock_meta_config, mock_common_config,
                  mock_proxy):
        hnv_metadata_agent.main()

        mock_config.register_agent_state_opts_helper.assert_called_once_with(
            CONF)
        mock_meta_config.register_meta_conf_opts.assert_called_once_with(
            hnv_metadata_agent.meta_config.METADATA_PROXY_HANDLER_OPTS)
        mock_common_config.init.assert_called_once_with(sys.argv[1:])
        mock_config.setup_logging.assert_called_once_with()
        mock_proxy.assert_called_once_with()
        mock_proxy.return_value.run.assert_called_once_with()



# Copyright (c) 2015 Cloudbase Solutions Srl
# Copyright (c) 2013 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import re

from neutron.plugins.ml2.drivers import mech_agent
from neutron_lib.api.definitions import portbindings

from networking_hyperv.neutron import constants


class HypervMechanismDriver(mech_agent.SimpleAgentMechanismDriverBase):
    """Attach to networks using Hyper-V L2 Agent.

    The HypervMechanismDriver integrates the Ml2 Plugin with the
    Hyperv L2 Agent. Port binding with this driver requires the Hyper-V
    agent to be running on the port's host, and that agent to have
    connectivity to at least one segment of the port's network.
    """

    def __init__(self):
        super(HypervMechanismDriver, self).__init__(
            constants.AGENT_TYPE_HYPERV,
            constants.VIF_TYPE_HYPERV,
            {portbindings.CAP_PORT_FILTER: False},
            supported_vnic_types=[portbindings.VNIC_NORMAL,
                                  portbindings.VNIC_DIRECT])

    def get_allowed_network_types(self, agent=None):
        network_types = [constants.TYPE_LOCAL, constants.TYPE_FLAT,
                         constants.TYPE_VLAN]
        if agent is not None:
            tunnel_types = agent.get('configurations', {}).get('tunnel_types')
            if tunnel_types:
                network_types.extend(tunnel_types)
        return network_types

    def get_mappings(self, agent):
        return agent['configurations'].get('vswitch_mappings', {})

    def physnet_in_mappings(self, physnet, mappings):
        return any(re.match(pattern, physnet) for pattern in mappings)



# Copyright 2017 Cloudbase Solutions SRL
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""This module contains all the available contract classes."""

import abc
import collections
import re

import eventlet
from eventlet import tpool
from neutron.agent import rpc as agent_rpc
from neutron_lib.agent import topics
from neutron_lib import constants as n_const
from neutron_lib import rpc as n_rpc
from os_win import constants as os_win_const
from os_win import exceptions as os_win_exc
from oslo_concurrency import lockutils
from oslo_log import log as logging
from oslo_service import loopingcall
import six

from networking_hyperv.common.i18n import _, _LI, _LE    # noqa
from networking_hyperv.neutron import _common_utils as c_util
from networking_hyperv.neutron.agent import base as base_agent
from networking_hyperv.neutron import config
from networking_hyperv.neutron import constants
from networking_hyperv.neutron import exception

LOG = logging.getLogger(__name__)
CONF = config.CONF

_synchronized = lockutils.synchronized_with_prefix('n-hv-agent-')
_port_synchronized = c_util.get_port_synchronized_decorator('n-hv-agent-')


class Layer2Agent(base_agent.BaseAgent):

    """Contract class for all the layer two agents."""

    _AGENT_TOPIC = n_const.L2_AGENT_TOPIC
    _OVS_EXT_NAME_RE = re.compile(r'.*((open.?v.?switch)|(ovs)).*',
                                  re.IGNORECASE)

    def __init__(self):
        super(Layer2Agent, self).__init__()
        self._network_vswitch_map = {}

        # The following sets contain ports that are to be processed.
        self._added_ports = set()
        self._removed_ports = set()

        # The following sets contain ports that have been processed.
        self._bound_ports = set()
        self._unbound_ports = set()

        self._physical_network_mappings = collections.OrderedDict()
        self._consumers = []
        self._event_callback_pairs = []

        # Setup the current agent.
        self._setup()
        self._set_agent_state()
        self._setup_rpc()

    def _setup(self):
        """Setup the layer two agent."""
        agent_config = CONF.get("AGENT", {})
        self._worker_count = agent_config.get('worker_count')
        self._phys_net_map = agent_config.get(
            'physical_network_vswitch_mappings', [])
        self._local_network_vswitch = agent_config.get(
            'local_network_vswitch')
        self._load_physical_network_mappings(self._phys_net_map)
        self._validate_vswitches()

        self._endpoints.append(self)
        self._event_callback_pairs.extend([
            (self._utils.EVENT_TYPE_CREATE, self._process_added_port_event),
            (self._utils.EVENT_TYPE_DELETE, self._process_removed_port_event)
        ])

        tpool.set_num_threads(self._worker_count)

    def _setup_qos_extension(self):
        """Setup the QOS extension if it is required."""
        pass

    def _setup_rpc(self):
        """Setup the RPC client for the current agent."""
        self._plugin_rpc = agent_rpc.PluginApi(topics.PLUGIN)
        self._state_rpc = agent_rpc.PluginReportStateAPI(topics.PLUGIN)
        self._client = n_rpc.get_client(self.target)

        self._consumers.extend([
            [topics.PORT, topics.UPDATE], [topics.NETWORK, topics.DELETE],
            [topics.PORT, topics.DELETE]
        ])

        self._connection = agent_rpc.create_consumers(
            self._endpoints, self._topic, self._consumers,
            start_listening=False
        )
        self._setup_qos_extension()
        self._connection.consume_in_threads()

        report_interval = CONF.AGENT.report_interval
        if report_interval:
            heartbeat = loopingcall.FixedIntervalLoopingCall(
                self._report_state)
            heartbeat.start(interval=report_interval)

    def _process_added_port_event(self, port_name):
        """Callback for added ports."""
        LOG.info("Hyper-V VM vNIC added: %s", port_name)
        self._added_ports.add(port_name)

    def _process_removed_port_event(self, port_name):
        LOG.info("Hyper-V VM vNIC removed: %s", port_name)
        self._removed_ports.add(port_name)

    def _load_physical_network_mappings(self, phys_net_vswitch_mappings):
        """Load all the information regarding the physical network."""
        for mapping in phys_net_vswitch_mappings:
            parts = mapping.split(':')
            if len(parts) != 2:
                LOG.debug('Invalid physical network mapping: %s', mapping)
            else:
                pattern = re.escape(parts[0].strip()).replace('\\*', '.*')
                pattern = pattern + '$'
                vswitch = parts[1].strip()
                self._physical_network_mappings[pattern] = vswitch

    def _validate_vswitches(self):
        vswitch_names = list(self._physical_network_mappings.values())
        if self._local_network_vswitch:
            vswitch_names.append(self._local_network_vswitch)

        vswitches_valid = True
        for vswitch_name in vswitch_names:
            try:
                self._validate_vswitch(vswitch_name)
            except exception.ValidationError:
                # We're validating all the vSwitches before erroring out.
                LOG.error("Validating vSwitch %s failed", vswitch_name)
                vswitches_valid = False

        # We're currently stopping the service if any of the configured
        # vSwitches are unavailable.
        if not vswitches_valid:
            err_msg = _("Validating one or more configured vSwitches failed.")
            raise exception.ValidationError(err_msg)
        elif not vswitch_names:
            err_msg = _("No vSwitch configured.")
            raise exception.ValidationError(err_msg)

    def _validate_vswitch(self, vswitch_name):
        try:
            vswitch_extensions = self._utils.get_vswitch_extensions(
                vswitch_name)
        except os_win_exc.HyperVvSwitchNotFound as exc:
            raise exception.ValidationError(exc.message)

        for ext in vswitch_extensions:
            if (self._is_ovs_extension(ext) and
                    ext['enabled_state'] == os_win_const.CIM_STATE_ENABLED):
                err_msg = _("The Open vSwitch extension is enabled on the "
                            "'%s' vSwitch. For this reason, this agent "
                            "cannot use the specified vSwitch.")
                raise exception.ValidationError(err_msg % vswitch_name)

    def _is_ovs_extension(self, vswitch_extension):
        # The OVS extension name keeps changing, while some vendors
        # redistribute it under a different name.
        return bool(self._OVS_EXT_NAME_RE.match(vswitch_extension['name']))

    def _get_vswitch_name(self, network_type, physical_network):
        """Get the vswitch name for the received network information."""
        if network_type != constants.TYPE_LOCAL:
            vswitch_name = self._get_vswitch_for_physical_network(
                physical_network)
        else:
            vswitch_name = self._local_network_vswitch

        if vswitch_name:
            return vswitch_name

        err_msg = _("No vSwitch configured for physical network "
                    "'%(physical_network)s'. Neutron network type: "
                    "'%(network_type)s'.")
        raise exception.NetworkingHyperVException(
            err_msg % dict(physical_network=physical_network,
                           network_type=network_type))

    def _get_vswitch_for_physical_network(self, phys_network_name):
        """Get the vswitch name for the received network name."""
        for pattern in self._physical_network_mappings:
            if phys_network_name is None:
                phys_network_name = ''
            if re.match(pattern, phys_network_name):
                return self._physical_network_mappings[pattern]

    def _get_network_vswitch_map_by_port_id(self, port_id):
        """Get the vswitch name for the received port id."""
        for network_id, vswitch in six.iteritems(self._network_vswitch_map):
            if port_id in vswitch['ports']:
                return (network_id, vswitch)

        # If the port was not found, just return (None, None)
        return (None, None)

    def _update_port_status_cache(self, device, device_bound=True):
        """Update the ports status cache."""
        with self._cache_lock:
            if device_bound:
                self._bound_ports.add(device)
                self._unbound_ports.discard(device)
            else:
                self._bound_ports.discard(device)
                self._unbound_ports.add(device)

    def _create_event_listeners(self):
        """Create and bind the event listeners."""
        LOG.debug("Create the event listeners.")
        for event_type, callback in self._event_callback_pairs:
            LOG.debug("Create listener for %r event", event_type)
            listener = self._utils.get_vnic_event_listener(event_type)
            eventlet.spawn_n(listener, callback)

    def _prologue(self):
        """Executed once before the daemon loop."""
        self._added_ports = self._utils.get_vnic_ids()
        self._create_event_listeners()

    def _reclaim_local_network(self, net_uuid):
        LOG.info("Reclaiming local network %s", net_uuid)
        del self._network_vswitch_map[net_uuid]

    def _port_bound(self, port_id, network_id, network_type, physical_network,
                    segmentation_id, port_security_enabled, set_port_sriov):
        """Bind the port to the recived network."""
        LOG.debug("Binding port %s", port_id)

        if network_id not in self._network_vswitch_map:
            self._provision_network(
                port_id, network_id, network_type,
                physical_network, segmentation_id)

        vswitch_map = self._network_vswitch_map[network_id]
        vswitch_map['ports'].append(port_id)

        LOG.debug("Trying to connect the current port to vswitch %r.",
                  vswitch_map['vswitch_name'])
        self._utils.connect_vnic_to_vswitch(
            vswitch_name=vswitch_map['vswitch_name'],
            switch_port_name=port_id,
        )
        if set_port_sriov:
            LOG.debug("Enabling SR-IOV for port: %s", port_id)
            self._utils.set_vswitch_port_sriov(port_id, True)

    def _port_unbound(self, port_id, vnic_deleted=False):
        LOG.debug("Trying to unbind the port %r", port_id)

        vswitch = self._get_network_vswitch_map_by_port_id(port_id)
        net_uuid, vswitch_map = vswitch

        if not net_uuid:
            LOG.debug('Port %s was not found on this agent.', port_id)
            return

        LOG.debug("Unbinding port %s", port_id)
        self._utils.remove_switch_port(port_id, vnic_deleted)
        vswitch_map['ports'].remove(port_id)

        if not vswitch_map['ports']:
            self._reclaim_local_network(net_uuid)

    def _process_added_port(self, device_details):
        # NOTE(claudiub): A port requiring SR-IOV will specify a PCI slot.
        set_port_sriov = 'pci_slot' in device_details.get('profile', {})
        self._treat_vif_port(
            port_id=device_details['port_id'],
            network_id=device_details['network_id'],
            network_type=device_details['network_type'],
            physical_network=device_details['physical_network'],
            segmentation_id=device_details['segmentation_id'],
            admin_state_up=device_details['admin_state_up'],
            port_security_enabled=device_details['port_security_enabled'],
            set_port_sriov=set_port_sriov)

    def process_added_port(self, device_details):
        """Process the new ports.

        Wraps _process_added_port, and treats the sucessful and exception
        cases.
        """
        device = device_details['device']
        port_id = device_details['port_id']
        reprocess = True
        try:
            self._process_added_port(device_details)

            LOG.debug("Updating cached port %s status as UP.", port_id)
            self._update_port_status_cache(device, device_bound=True)
            LOG.info("Port %s processed.", port_id)
        except os_win_exc.HyperVvNicNotFound:
            LOG.debug('vNIC %s not found. This can happen if the VM was '
                      'destroyed.', port_id)
            reprocess = False
        except os_win_exc.HyperVPortNotFoundException:
            LOG.debug('vSwitch port %s not found. This can happen if the VM '
                      'was destroyed.', port_id)
            # NOTE(claudiub): just to be on the safe side, in case Hyper-V said
            # that the port was added, but it hasn't really, we're leaving
            # reprocess = True. If the VM / vNIC was removed, on the next
            # reprocess, a HyperVvNicNotFound will be raised.
        except Exception as ex:
            # NOTE(claudiub): in case of a non-transient error, the port will
            # be processed over and over again, and will not be reported as
            # bound (e.g.: InvalidParameterValue when setting QoS), until the
            # port is deleted. These issues have to be investigated and solved
            LOG.exception("Exception encountered while processing "
                          "port %(port_id)s. Exception: %(ex)s",
                          dict(port_id=port_id, ex=ex))
        else:
            # no exception encountered, no need to reprocess.
            reprocess = False

        if reprocess:
            # Readd the port as "added", so it can be reprocessed.
            self._added_ports.add(device)

            # Force cache refresh.
            self._refresh_cache = True
            return False

        return True

    def _treat_devices_added(self):
        """Process the new devices."""
        try:
            devices_details_list = self._plugin_rpc.get_devices_details_list(
                self._context, self._added_ports, self._agent_id, self._host)
        except Exception as exc:
            LOG.debug("Unable to get ports details for "
                      "devices %(devices)s: %(exc)s",
                      {'devices': self._added_ports, 'exc': exc})
            return

        for device_details in devices_details_list:
            device = device_details['device']
            LOG.info("Adding port %s", device)
            if 'port_id' in device_details:
                LOG.info("Port %(device)s updated. "
                         "Details: %(device_details)s",
                         {'device': device, 'device_details': device_details})
                eventlet.spawn_n(self.process_added_port, device_details)
            else:
                LOG.debug("Missing port_id from device details: "
                          "%(device)s. Details: %(device_details)s",
                          {'device': device, 'device_details': device_details})

            LOG.debug("Remove the port from added ports set, so it "
                      "doesn't get reprocessed.")
            self._added_ports.discard(device)

    def _process_removed_port(self, device):
        """Process the removed ports."""
        LOG.debug("Trying to remove the port %r", device)
        self._update_port_status_cache(device, device_bound=False)
        self._port_unbound(device, vnic_deleted=True)

        LOG.debug("The port was successfully removed.")
        self._removed_ports.discard(device)

    def _treat_devices_removed(self):
        """Process the removed devices."""
        for device in self._removed_ports.copy():
            eventlet.spawn_n(self._process_removed_port, device)

    @_synchronized('n-plugin-notifier')
    def _notify_plugin_on_port_updates(self):
        if not (self._bound_ports or self._unbound_ports):
            return

        with self._cache_lock:
            bound_ports = self._bound_ports.copy()
            unbound_ports = self._unbound_ports.copy()

        self._plugin_rpc.update_device_list(
            self._context, list(bound_ports), list(unbound_ports),
            self._agent_id, self._host)

        with self._cache_lock:
            self._bound_ports = self._bound_ports.difference(bound_ports)
            self._unbound_ports = self._unbound_ports.difference(
                unbound_ports)

    def _work(self):
        """Process the information regarding the available ports."""
        if self._refresh_cache:
            # Inconsistent cache might cause exceptions. For example,
            # if a port has been removed, it will be known in the next
            # loop. Using the old switch port can cause exceptions.
            LOG.debug("Refreshing os_win caches...")
            self._utils.update_cache()
            self._refresh_cache = False

        if self._bound_ports or self._unbound_ports:
            eventlet.spawn_n(self._notify_plugin_on_port_updates)

        # notify plugin about port deltas
        if self._added_ports:
            LOG.debug("Agent loop has new devices!")
            self._treat_devices_added()

        if self._removed_ports:
            LOG.debug("Agent loop has lost devices...")
            self._treat_devices_removed()

    def port_update(self, context, port=None, network_type=None,
                    segmentation_id=None, physical_network=None):
        LOG.debug("port_update received: %s", port['id'])

        if self._utils.vnic_port_exists(port['id']):
            self._treat_vif_port(
                port_id=port['id'],
                network_id=port['network_id'],
                network_type=network_type,
                physical_network=physical_network,
                segmentation_id=segmentation_id,
                admin_state_up=port['admin_state_up'],
                port_security_enabled=port['port_security_enabled'],
            )
        else:
            LOG.debug("No port %s defined on agent.", port['id'])

    def port_delete(self, context, port_id=None):
        """Delete the received port."""
        LOG.debug("port_delete event received for %r", port_id)

    def network_delete(self, context, network_id=None):
        LOG.debug("network_delete received. "
                  "Deleting network %s", network_id)

        # The network may not be defined on this agent
        if network_id in self._network_vswitch_map:
            self._reclaim_local_network(network_id)
        else:
            LOG.debug("Network %s not defined on agent.", network_id)

    @abc.abstractmethod
    def _provision_network(self, port_id, net_uuid, network_type,
                           physical_network, segmentation_id):
        """Provision the network with the received information."""
        pass

    @_port_synchronized
    def _treat_vif_port(self, port_id, network_id, network_type,
                        physical_network, segmentation_id,
                        admin_state_up, port_security_enabled,
                        set_port_sriov=False):
        if admin_state_up:
            self._port_bound(port_id, network_id, network_type,
                             physical_network, segmentation_id,
                             port_security_enabled, set_port_sriov)
        else:
            self._port_unbound(port_id)



# Copyright 2017 Cloudbase Solutions Srl
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Unit tests for the networking-hyperv config module.
"""

import mock

from networking_hyperv.neutron import config
from networking_hyperv.tests import base


class TestConfig(base.HyperVBaseTestCase):

    @mock.patch.object(config, 'ks_loading')
    @mock.patch.object(config, 'CONF')
    def test_register_opts(self, mock_CONF, mock_ks_loading):
        config.register_opts()

        all_groups = [pair[0] for pair in config.ALL_OPTS]
        mock_CONF.register_group.assert_has_calls([
            mock.call(group) for group in all_groups])

        mock_CONF.register_opts.assert_has_calls([
            mock.call(opts, group=group) for group, opts in config.ALL_OPTS])

        mock_ks_loading.register_session_conf_options.assert_called_once_with(
            mock_CONF, config.NEUTRON_GROUP)
        mock_ks_loading.register_auth_conf_options.assert_called_once_with(
            mock_CONF, config.NEUTRON_GROUP)



# Copyright 2015 Cloudbase Solutions SRL
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from keystoneauth1 import loading as ks_loading
from neutronclient.v2_0 import client as clientv20
from oslo_log import log as logging

from networking_hyperv.common.i18n import _LW, _LE  # noqa
from networking_hyperv.neutron import config
from networking_hyperv.neutron import constants

CONF = config.CONF
LOG = logging.getLogger(__name__)


class NeutronAPIClient(object):

    def __init__(self):
        self._init_client()

    def _init_client(self):
        session = ks_loading.load_session_from_conf_options(
            CONF, config.NEUTRON_GROUP)
        auth_plugin = ks_loading.load_auth_from_conf_options(
            CONF, config.NEUTRON_GROUP)

        self._client = clientv20.Client(
            session=session,
            auth=auth_plugin)

    def get_network_subnets(self, network_id):
        try:
            net = self._client.show_network(network_id)
            return net['network']['subnets']
        except Exception as ex:
            LOG.error("Could not retrieve network %(network_id)s . Error: "
                      "%(ex)s", {'network_id': network_id, 'ex': ex})

        return []

    def get_network_subnet_cidr_and_gateway(self, subnet_id):
        try:
            subnet = self._client.show_subnet(subnet_id)['subnet']
            return (str(subnet['cidr']), str(subnet['gateway_ip']))
        except Exception as ex:
            LOG.error("Could not retrieve subnet %(subnet_id)s . Error: "
                      "%(ex)s: ", {'subnet_id': subnet_id, 'ex': ex})

        return None, None

    def get_port_ip_address(self, port_id):
        try:
            port = self._client.show_port(port_id)
            fixed_ips = port['port']['fixed_ips'][0]
            return fixed_ips['ip_address']
        except Exception as ex:
            LOG.error("Could not retrieve port %(port_id)s . Error: "
                      "%(ex)s", {'port_id': port_id, 'ex': ex})

        return None

    def get_tunneling_agents(self):
        try:
            agents = self._client.list_agents()
            tunneling_agents = [
                a for a in agents['agents'] if constants.TYPE_NVGRE in
                a.get('configurations', {}).get('tunnel_types', [])]

            tunneling_ip_agents = [
                a for a in tunneling_agents if
                a.get('configurations', {}).get('tunneling_ip')]

            if len(tunneling_ip_agents) < len(tunneling_agents):
                LOG.warning('Some agents have NVGRE tunneling enabled, but'
                            ' do not provide tunneling_ip. Ignoring those '
                            'agents.')

            return dict([(a['host'], a['configurations']['tunneling_ip'])
                         for a in tunneling_ip_agents])
        except Exception as ex:
            LOG.error("Could not get tunneling agents. Error: %s", ex)
            return {}

    def get_network_ports(self, **kwargs):
        try:
            return self._client.list_ports(**kwargs)['ports']
        except Exception as ex:
            LOG.error("Exception caught: %s", ex)
        return []

    def get_port_profile_id(self, port_id):
        try:
            port = self._client.show_port(port_id)
            return "{%s}" % (port["port"]["binding:vif_details"]
                             ["port_profile_id"])
        except Exception:
            LOG.exception("Failed to retrieve profile id for port %s.",
                          port_id)
        return {}



# Copyright 2013 Cloudbase Solutions SRL
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

# Topic for tunnel notifications between the plugin and agent
AGENT_TOPIC = 'q-agent-notifier'
AGENT_TYPE_HYPERV = 'HyperV agent'
AGENT_TYPE_HNV = "HNV agent"
VIF_TYPE_HYPERV = 'hyperv'

TUNNEL = 'tunnel'
LOOKUP = 'lookup'

UPDATE = 'update'

# Special vlan_id value in ovs_vlan_allocations table indicating flat network
FLAT_VLAN_ID = -1

TYPE_FLAT = 'flat'
TYPE_LOCAL = 'local'
TYPE_VLAN = 'vlan'
TYPE_NVGRE = 'gre'

IPV4_DEFAULT = '0.0.0.0'

# Windows Server 2016 Network Controller related constants.
# NOTE(claudiub): These constants HAVE to be defined exactly like this,
# otherwise networking using the Windows Server 2016 Network Controller won't
# work.
# https://docs.microsoft.com/en-us/windows-server/networking/sdn/manage/create-a-tenant-vm  # noqa
NET_CFG_INSTANCE_ID = "{56785678-a0e5-4a26-bc9b-c0cba27311a3}"
CDN_LABEL_STRING = "OpenStackCdn"
CDN_LABEL_ID = 1111
PROFILE_NAME = "OpenStackProfile"
VENDOR_ID = "{1FA41B39-B444-4E43-B35A-E1F7985FD548}"
VENDOR_NAME = "NetworkController"
PROFILE_DATA = 1



# Copyright 2014 Cloudbase Solutions SRL
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Unit tests for the Hyper-V Security Groups Driver.
"""

import mock
from os_win import exceptions

from networking_hyperv.neutron import security_groups_driver as sg_driver
from networking_hyperv.tests import base


class SecurityGroupRuleTestHelper(base.HyperVBaseTestCase):
    _FAKE_DIRECTION = 'egress'
    _FAKE_ETHERTYPE = 'IPv4'
    _FAKE_ETHERTYPE_IPV6 = 'IPv6'
    _FAKE_PROTOCOL = 'tcp'
    _FAKE_ACTION = sg_driver.ACL_PROP_MAP['action']['allow']
    _FAKE_DEST_IP_PREFIX = '10.0.0.0/24'
    _FAKE_SOURCE_IP_PREFIX = '10.0.1.0/24'
    _FAKE_MEMBER_IP = '10.0.0.1'
    _FAKE_IPV6_LEN128_IP = 'fddd:cafd:e664:0:f816:3eff:fe8d:59d2/128'
    _FAKE_SG_ID = 'fake_sg_id'

    _FAKE_PORT_MIN = 9001
    _FAKE_PORT_MAX = 9011

    def _create_security_rule(self, **rule_updates):
        rule = {
            'direction': self._FAKE_DIRECTION,
            'ethertype': self._FAKE_ETHERTYPE,
            'protocol': self._FAKE_PROTOCOL,
            'dest_ip_prefix': self._FAKE_DEST_IP_PREFIX,
            'source_ip_prefix': self._FAKE_SOURCE_IP_PREFIX,
            'port_range_min': self._FAKE_PORT_MIN,
            'port_range_max': self._FAKE_PORT_MAX,
            'security_group_id': self._FAKE_SG_ID
        }
        rule.update(rule_updates)
        return rule

    @classmethod
    def _acl(self, key1, key2):
        return sg_driver.ACL_PROP_MAP[key1][key2]


class TestHyperVSecurityGroupsDriver(SecurityGroupRuleTestHelper):

    _autospec_classes = [
        sg_driver.SecurityGroupRuleGeneratorR2,
    ]

    _FAKE_DEVICE = 'fake_device'
    _FAKE_ID = 'fake_id'
    _FAKE_PARAM_NAME = 'fake_param_name'
    _FAKE_PARAM_VALUE = 'fake_param_value'

    def setUp(self):
        super(TestHyperVSecurityGroupsDriver, self).setUp()

        self._driver = sg_driver.HyperVSecurityGroupsDriver()
        self._driver._utils = mock.MagicMock(autospec=self._driver._utils)

    def test__select_sg_rules_for_port(self):
        mock_port = self._get_port()
        mock_port['fixed_ips'] = [mock.MagicMock()]
        mock_port['security_groups'] = [self._FAKE_SG_ID]

        fake_sg_template = self._create_security_rule()
        fake_sg_template['direction'] = 'ingress'
        self._driver._sg_rule_templates[self._FAKE_SG_ID] = [fake_sg_template]

        # Test without remote_group_id
        rule_list = self._driver._select_sg_rules_for_port(mock_port,
                                                           'ingress')
        self.assertNotIn('security_group_id', rule_list[0])

        # Test with remote_group_id
        fake_sg_template['remote_group_id'] = self._FAKE_SG_ID
        self._driver._sg_members[self._FAKE_SG_ID] = {self._FAKE_ETHERTYPE:
                                                      [self._FAKE_MEMBER_IP]}
        rule_list = self._driver._select_sg_rules_for_port(mock_port,
                                                           'ingress')
        self.assertEqual('10.0.0.1/32', rule_list[0]['source_ip_prefix'])
        self.assertNotIn('security_group_id', rule_list[0])
        self.assertNotIn('remote_group_id', rule_list[0])

        # Test for fixed 'ip' existing in 'sg_members'
        self._driver._sg_members[self._FAKE_SG_ID][self._FAKE_ETHERTYPE] = [
            '10.0.0.2']

        mock_port['fixed_ips'] = ['10.0.0.2']
        rule_list = self._driver._select_sg_rules_for_port(mock_port,
                                                           'ingress')
        self.assertEqual([], rule_list)

        # Test for 'egress' direction
        fake_sg_template['direction'] = 'egress'
        fix_ip = [self._FAKE_MEMBER_IP, '10.0.0.2']
        self._driver._sg_members[self._FAKE_SG_ID][self._FAKE_ETHERTYPE] = (
            fix_ip)

        rule_list = self._driver._select_sg_rules_for_port(mock_port,
                                                           'egress')
        self.assertEqual('10.0.0.1/32', rule_list[0]['dest_ip_prefix'])

        # Test for rules with a different direction
        rule_list = self._driver._select_sg_rules_for_port(mock_port,
                                                           'ingress')
        self.assertEqual([], rule_list)

    def test_update_security_group_rules(self):
        mock_rule = [self._create_security_rule()]
        self._driver.update_security_group_rules(self._FAKE_ID, mock_rule)
        self.assertEqual(mock_rule,
                         self._driver._sg_rule_templates[self._FAKE_ID])

    def test_update_security_group_members(self):
        mock_member = ['10.0.0.1/32']
        self._driver.update_security_group_members(self._FAKE_ID, mock_member)
        self.assertEqual(mock_member, self._driver._sg_members[self._FAKE_ID])

    @mock.patch.object(sg_driver.HyperVSecurityGroupsDriver,
                       '_select_sg_rules_for_port')
    def test__generate_rules(self, mock_select_sg_rules):
        mock_rule = [self._create_security_rule()]
        mock_port = self._get_port()
        mock_select_sg_rules.return_value = mock_rule
        ports = self._driver._generate_rules([mock_port])

        # Expected result
        mock_rule.append(mock_rule[0])
        expected = {self._FAKE_ID: mock_rule}
        self.assertEqual(expected, ports)

    @mock.patch.object(sg_driver.HyperVSecurityGroupsDriver,
                       '_generate_rules')
    @mock.patch.object(sg_driver.HyperVSecurityGroupsDriver,
                       '_create_port_rules')
    @mock.patch.object(sg_driver.HyperVSecurityGroupsDriver,
                       '_add_sg_port_rules')
    def test_prepare_port_filter(self, mock_add_rules, mock_create_rules,
                                 mock_gen_rules):
        mock_port = self._get_port()
        mock_create_default = self._driver._sg_gen.create_default_sg_rules

        fake_rule = self._create_security_rule()
        self._driver._get_rule_remote_address = mock.MagicMock(
            return_value=self._FAKE_SOURCE_IP_PREFIX)
        mock_gen_rules.return_value = {mock_port['id']: [fake_rule]}

        self._driver.prepare_port_filter(mock_port)

        self.assertEqual(mock_port,
                         self._driver._security_ports[self._FAKE_DEVICE])
        mock_gen_rules.assert_called_with([self._driver._security_ports
                                          [self._FAKE_DEVICE]])
        mock_add_rules.assert_called_once_with(
            mock_port, mock_create_default.return_value)

        self._driver._create_port_rules.assert_called_with(
            mock_port, [fake_rule])

    def test_prepare_port_filter_security_disabled(self):
        mock_port = self._get_port()
        mock_port.pop('port_security_enabled')

        self._driver.prepare_port_filter(mock_port)

        self.assertNotIn(mock_port['device'], self._driver._security_ports)
        self.assertNotIn(mock_port['id'], self._driver._sec_group_rules)

    @mock.patch.object(sg_driver.HyperVSecurityGroupsDriver,
                       '_generate_rules')
    def test_update_port_filter(self, mock_gen_rules):
        mock_port = self._get_port()
        new_mock_port = self._get_port()
        new_mock_port['id'] += '2'
        new_mock_port['security_group_rules'][0]['ethertype'] += "2"

        fake_rule_new = self._create_security_rule()
        self._driver._get_rule_remote_address = mock.MagicMock(
            return_value=self._FAKE_SOURCE_IP_PREFIX)

        mock_gen_rules.return_value = {new_mock_port['id']: [fake_rule_new]}
        self._driver._sg_gen.expand_wildcard_rules.return_value = []

        self._driver._security_ports[mock_port['device']] = mock_port
        self._driver._sec_group_rules[new_mock_port['id']] = []

        self._driver._create_port_rules = mock.MagicMock()
        self._driver._remove_port_rules = mock.MagicMock()
        self._driver.update_port_filter(new_mock_port)

        self._driver._remove_port_rules.assert_called_once_with(
            mock_port, mock_port['security_group_rules'])
        self._driver._create_port_rules.assert_called_once_with(
            new_mock_port, [new_mock_port['security_group_rules'][0],
                            fake_rule_new])
        self.assertEqual(new_mock_port,
                         self._driver._security_ports[new_mock_port['device']])

    @mock.patch.object(sg_driver.HyperVSecurityGroupsDriver,
                       'prepare_port_filter')
    def test_update_port_filter_new_port(self, mock_method):
        mock_port = self._get_port()
        new_mock_port = self._get_port()
        new_mock_port['id'] += '2'
        new_mock_port['device'] += '2'
        new_mock_port['security_group_rules'][0]['ethertype'] += "2"

        self._driver._security_ports[mock_port['device']] = mock_port
        self._driver.update_port_filter(new_mock_port)

        self.assertNotIn(new_mock_port['device'], self._driver._security_ports)
        mock_method.assert_called_once_with(new_mock_port)

    def test_update_port_filter_security_disabled_existing_rules(self):
        mock_port = self._get_port()
        mock_port.pop('port_security_enabled')
        self._driver._sec_group_rules[mock_port['id']] = mock.ANY
        self._driver._security_ports[mock_port['device']] = mock.sentinel.port

        self._driver.update_port_filter(mock_port)

        self._driver._utils.remove_all_security_rules.assert_called_once_with(
            mock_port['id'])
        self.assertNotIn(mock_port['device'], self._driver._security_ports)
        self.assertNotIn(mock_port['id'], self._driver._sec_group_rules)

    @mock.patch.object(sg_driver.HyperVSecurityGroupsDriver,
                       '_generate_rules')
    def test_update_port_filter_existing_wildcard_rules(self, mock_gen_rules):
        mock_port = self._get_port()
        new_mock_port = self._get_port()
        new_mock_port['id'] += '2'
        new_mock_port['security_group_rules'][0]['ethertype'] += "2"

        fake_rule_new = self._create_security_rule(direction='egress',
                                                   protocol='udp')
        fake_wildcard_rule_new = self._create_security_rule(protocol='ANY',
                                                            direction='egress')
        fake_expanded_rules = [
            self._create_security_rule(direction='egress', protocol='tcp'),
            self._create_security_rule(direction='egress', protocol='udp'),
            self._create_security_rule(direction='egress', protocol='icmp')]
        self._driver._sg_gen.expand_wildcard_rules.return_value = (
            fake_expanded_rules)

        mock_gen_rules.return_value = {
            new_mock_port['id']: [fake_rule_new, fake_wildcard_rule_new]}

        self._driver._security_ports[mock_port['device']] = mock_port
        self._driver._sec_group_rules[new_mock_port['id']] = [
            self._create_security_rule(direction='egress', protocol='icmp')]

        filtered_new_rules = [new_mock_port['security_group_rules'][0],
                              fake_wildcard_rule_new]
        filtered_remove_rules = mock_port['security_group_rules']

        self._driver._create_port_rules = mock.MagicMock()
        self._driver._remove_port_rules = mock.MagicMock()
        self._driver.update_port_filter(new_mock_port)

        self._driver._sg_gen.expand_wildcard_rules.assert_called_once_with(
            [fake_rule_new, fake_wildcard_rule_new])
        self._driver._remove_port_rules.assert_called_once_with(
            mock_port, filtered_remove_rules)
        self._driver._create_port_rules.assert_called_once_with(
            new_mock_port, filtered_new_rules)
        self.assertEqual(new_mock_port,
                         self._driver._security_ports[new_mock_port['device']])

    def test_remove_port_filter(self):
        mock_port = self._get_port()
        mock_rule = mock.MagicMock()
        self._driver._sec_group_rules[self._FAKE_ID] = [mock_rule]
        self._driver._security_ports[mock_port['device']] = mock_port
        self._driver.remove_port_filter(mock_port)
        self.assertNotIn(mock_port['device'], self._driver._security_ports)
        self.assertNotIn(mock_port['id'], self._driver._sec_group_rules)
        self._driver._utils.clear_port_sg_acls_cache(mock_port['id'])

    @mock.patch.object(sg_driver.HyperVSecurityGroupsDriver,
                       '_add_sg_port_rules')
    @mock.patch.object(sg_driver.HyperVSecurityGroupsDriver,
                       '_remove_sg_port_rules')
    def test_create_port_rules(self, mock_remove, mock_add):
        mock_port = self._get_port()
        mock_rule = mock.MagicMock()
        self._driver._sec_group_rules[self._FAKE_ID] = [mock_rule]
        self._driver._sg_gen.create_security_group_rules.return_value = [
            mock_rule]
        self._driver._sg_gen.compute_new_rules_add.return_value = (
            [mock_rule, mock_rule], [mock_rule, mock_rule])

        self._driver._create_port_rules(mock_port, [mock_rule])

        self._driver._sg_gen.compute_new_rules_add.assert_called_once_with(
            [mock_rule], [mock_rule])
        mock_remove.assert_called_once_with(mock_port, [mock_rule])
        mock_add.assert_called_once_with(mock_port, [mock_rule])

    @mock.patch.object(sg_driver.HyperVSecurityGroupsDriver,
                       '_remove_sg_port_rules')
    def test_remove_port_rules(self, mock_remove):
        mock_port = self._get_port()
        mock_rule = mock.MagicMock()
        self._driver._sec_group_rules[self._FAKE_ID] = [mock_rule]
        self._driver._sg_gen.create_security_group_rules.return_value = [
            mock_rule]

        self._driver._remove_port_rules(mock_port, [mock_rule])

        mock_remove.assert_called_once_with(mock_port, [mock_rule])

    def test_add_sg_port_rules_exception(self):
        mock_port = self._get_port()
        mock_rule = mock.MagicMock()
        self._driver._sec_group_rules[self._FAKE_ID] = []
        self._driver._utils.create_security_rules.side_effect = (
            exceptions.HyperVException(msg='Generated Exception for testing.'))

        self.assertRaises(exceptions.HyperVException,
                          self._driver._add_sg_port_rules,
                          mock_port, [mock_rule])

        self.assertNotIn(mock_rule,
                         self._driver._sec_group_rules[self._FAKE_ID])

    def test_add_sg_port_rules_port_not_found(self):
        mock_port = self._get_port()
        self._driver._sec_group_rules[self._FAKE_ID] = []
        self._driver._security_ports[self._FAKE_DEVICE] = mock.sentinel.port
        self._driver._utils.create_security_rules.side_effect = (
            exceptions.NotFound(resource='port_id'))

        self.assertRaises(exceptions.NotFound,
                          self._driver._add_sg_port_rules,
                          mock_port, [mock.sentinel.rule])

        self.assertNotIn(self._FAKE_ID, self._driver._sec_group_rules)
        self.assertNotIn(self._FAKE_DEVICE, self._driver._security_ports)

    def test_add_sg_port_rules(self):
        mock_port = self._get_port()
        mock_rule = mock.MagicMock()
        self._driver._sec_group_rules[self._FAKE_ID] = []
        self._driver._add_sg_port_rules(mock_port, [mock_rule])

        self._driver._utils.create_security_rules.assert_called_once_with(
            self._FAKE_ID, [mock_rule])
        self.assertIn(mock_rule, self._driver._sec_group_rules[self._FAKE_ID])

    def test_add_sg_port_rules_empty(self):
        mock_port = self._get_port()
        self._driver._add_sg_port_rules(mock_port, [])
        self.assertFalse(self._driver._utils.create_security_rules.called)

    def test_remove_sg_port_rules_exception(self):
        mock_port = self._get_port()
        mock_rule = mock.MagicMock()
        self._driver._sec_group_rules[self._FAKE_ID] = [mock_rule]
        self._driver._utils.remove_security_rules.side_effect = (
            exceptions.HyperVException(msg='Generated Exception for testing.'))

        self.assertRaises(exceptions.HyperVException,
                          self._driver._remove_sg_port_rules,
                          mock_port, [mock_rule])

        self.assertIn(mock_rule, self._driver._sec_group_rules[self._FAKE_ID])

    def test_remove_sg_port_rules_port_not_found(self):
        mock_port = self._get_port()
        self._driver._sec_group_rules[self._FAKE_ID] = []
        self._driver._security_ports[self._FAKE_DEVICE] = mock.sentinel.port
        self._driver._utils.remove_security_rules.side_effect = (
            exceptions.NotFound(resource='port_id'))

        self.assertRaises(exceptions.NotFound,
                          self._driver._remove_sg_port_rules,
                          mock_port, [mock.sentinel.rule])

        self.assertNotIn(self._FAKE_ID, self._driver._sec_group_rules)
        self.assertNotIn(self._FAKE_DEVICE, self._driver._security_ports)

    def test_remove_sg_port_rules(self):
        mock_port = self._get_port()
        mock_rule = mock.MagicMock()
        self._driver._sec_group_rules[self._FAKE_ID] = [mock_rule]
        self._driver._remove_sg_port_rules(
            mock_port, [mock_rule, mock.sentinel.other_rule])

        self._driver._utils.remove_security_rules.assert_called_once_with(
            self._FAKE_ID, [mock_rule, mock.sentinel.other_rule])
        self.assertNotIn(mock_rule,
                         self._driver._sec_group_rules[self._FAKE_ID])

    def test_remove_sg_port_rules_empty(self):
        mock_port = self._get_port()
        self._driver._remove_sg_port_rules(mock_port, [])
        self.assertFalse(self._driver._utils.remove_security_rules.called)

    def _get_port(self):
        return {
            'device': self._FAKE_DEVICE,
            'id': self._FAKE_ID,
            'security_group_rules': [mock.MagicMock()],
            'port_security_enabled': True
        }


class SecurityGroupRuleR2BaseTestCase(SecurityGroupRuleTestHelper):
    def _create_sg_rule(self, protocol=None, action=None, direction='egress'):
        protocol = protocol or self._FAKE_PROTOCOL
        action = action or self._FAKE_ACTION
        remote_addr = (self._FAKE_DEST_IP_PREFIX if direction is 'egress' else
                       self._FAKE_SOURCE_IP_PREFIX)
        return sg_driver.SecurityGroupRuleR2(
            self._acl('direction', self._FAKE_DIRECTION),
            '%s-%s' % (self._FAKE_PORT_MIN, self._FAKE_PORT_MAX),
            protocol, remote_addr, action)


class SecurityGroupRuleGeneratorTestCase(SecurityGroupRuleR2BaseTestCase):

    def setUp(self):
        super(SecurityGroupRuleGeneratorTestCase, self).setUp()

        self.sg_gen = sg_driver.SecurityGroupRuleGenerator()

    @mock.patch.object(sg_driver.SecurityGroupRuleGenerator,
                       'create_security_group_rule')
    def test_create_security_group_rules(self, mock_create_sec_group_rule):
        sg_rule = self._create_sg_rule()
        mock_create_sec_group_rule.return_value = [sg_rule]
        expected = [sg_rule] * 2
        rules = [self._create_security_rule()] * 2

        actual = self.sg_gen.create_security_group_rules(rules)
        self.assertEqual(expected, actual)

    def test_convert_any_address_to_same_ingress(self):
        rule = self._create_security_rule()
        rule['direction'] = 'ingress'
        actual = self.sg_gen._get_rule_remote_address(rule)
        self.assertEqual(self._FAKE_SOURCE_IP_PREFIX, actual)

    def test_convert_any_address_to_same_egress(self):
        rule = self._create_security_rule()
        rule['direction'] += '2'
        actual = self.sg_gen._get_rule_remote_address(rule)
        self.assertEqual(self._FAKE_DEST_IP_PREFIX, actual)

    def test_convert_any_address_to_ipv4(self):
        rule = self._create_security_rule()
        del rule['dest_ip_prefix']
        actual = self.sg_gen._get_rule_remote_address(rule)
        self.assertEqual(self._acl('address_default', 'IPv4'), actual)

    def test_convert_any_address_to_ipv6(self):
        rule = self._create_security_rule()
        del rule['dest_ip_prefix']
        rule['ethertype'] = self._FAKE_ETHERTYPE_IPV6
        actual = self.sg_gen._get_rule_remote_address(rule)
        self.assertEqual(self._acl('address_default', 'IPv6'), actual)


class SecurityGroupRuleGeneratorR2TestCase(SecurityGroupRuleR2BaseTestCase):

    def setUp(self):
        super(SecurityGroupRuleGeneratorR2TestCase, self).setUp()

        self.sg_gen = sg_driver.SecurityGroupRuleGeneratorR2()

    def test_create_security_group_rule(self):
        expected = [self._create_sg_rule()]
        rule = self._create_security_rule()

        actual = self.sg_gen.create_security_group_rule(rule)
        self.assertEqual(expected, actual)

    def test_create_security_group_rule_len128(self):
        expected = [self._create_sg_rule()]
        expected[0].RemoteIPAddress = self._FAKE_IPV6_LEN128_IP.split(
            '/128', 1)[0]
        rule = self._create_security_rule()
        rule['dest_ip_prefix'] = self._FAKE_IPV6_LEN128_IP

        actual = self.sg_gen.create_security_group_rule(rule)
        self.assertEqual(expected, actual)

    def test_create_security_group_rule_any(self):
        sg_rule1 = self._create_sg_rule(self._acl('protocol', 'tcp'))
        sg_rule2 = self._create_sg_rule(self._acl('protocol', 'udp'))
        sg_rule3 = self._create_sg_rule(self._acl('protocol', 'icmp'))
        sg_rule4 = self._create_sg_rule(self._acl('protocol', 'ipv6-icmp'))

        rule = self._create_security_rule()
        rule['protocol'] = sg_driver.ACL_PROP_MAP["default"]

        actual = self.sg_gen.create_security_group_rule(rule)
        expected = [sg_rule1, sg_rule2, sg_rule3, sg_rule4]
        self.assertEqual(sorted(expected), sorted(actual))

    def test_create_default_sg_rules(self):
        actual = self.sg_gen.create_default_sg_rules()
        self.assertEqual(16, len(actual))

    def test_compute_new_rules_add(self):
        new_rule = self._create_sg_rule()
        old_rule = self._create_sg_rule()
        old_rule.Direction = mock.sentinel.FAKE_DIRECTION

        add_rules, remove_rules = self.sg_gen.compute_new_rules_add(
            [old_rule], [new_rule, old_rule])

        self.assertEqual([new_rule], add_rules)

    def test_expand_wildcard_rules(self):
        egress_wildcard_rule = self._create_security_rule(
            protocol='ANY',
            direction='egress')
        ingress_wildcard_rule = self._create_security_rule(
            protocol='ANY',
            direction='ingress')
        normal_rule = self._create_security_rule()
        rules = [egress_wildcard_rule, ingress_wildcard_rule, normal_rule]

        actual_expanded_rules = self.sg_gen.expand_wildcard_rules(rules)

        expanded_rules = []
        for proto in sg_driver.ACL_PROP_MAP['protocol'].keys():
            expanded_rules.extend(
                [self._create_security_rule(protocol=proto,
                                            direction='egress'),
                 self._create_security_rule(protocol=proto,
                                            direction='ingress')])
        diff_expanded_rules = [r for r in expanded_rules
                               if r not in actual_expanded_rules]
        self.assertEqual([], diff_expanded_rules)

    def test_expand_no_wildcard_rules(self):
        normal_rule = self._create_security_rule(direction='egress')
        another_normal_rule = self._create_security_rule(direction='ingress')

        actual_expanded_rules = self.sg_gen.expand_wildcard_rules(
            [normal_rule, another_normal_rule])

        self.assertEqual([], actual_expanded_rules)

    def test_get_rule_port_range(self):
        rule = self._create_security_rule()
        expected = '%s-%s' % (self._FAKE_PORT_MIN, self._FAKE_PORT_MAX)
        actual = self.sg_gen._get_rule_port_range(rule)

        self.assertEqual(expected, actual)

    def test_get_rule_port_range_default(self):
        rule = self._create_security_rule()
        del rule['port_range_min']
        expected = sg_driver.ACL_PROP_MAP['default']
        actual = self.sg_gen._get_rule_port_range(rule)

        self.assertEqual(expected, actual)

    def test_get_rule_protocol_icmp_ipv6(self):
        self._check_get_rule_protocol(
            expected=self._acl('protocol', 'ipv6-icmp'),
            protocol='icmp',
            ethertype='IPv6')

    def test_get_rule_protocol_icmp(self):
        self._check_get_rule_protocol(expected=self._acl('protocol', 'icmp'),
                                      protocol='icmp')

    def test_get_rule_protocol_no_icmp(self):
        self._check_get_rule_protocol(expected='tcp',
                                      protocol='tcp')

    def _check_get_rule_protocol(self, expected, **rule_updates):
        rule = self._create_security_rule(**rule_updates)
        actual = self.sg_gen._get_rule_protocol(rule)

        self.assertEqual(expected, actual)


class SecurityGroupRuleR2TestCase(SecurityGroupRuleR2BaseTestCase):

    def test_sg_rule_to_dict(self):
        expected = {'Direction': self._acl('direction', self._FAKE_DIRECTION),
                    'Action': self._FAKE_ACTION,
                    'Protocol': self._FAKE_PROTOCOL,
                    'LocalPort': '%s-%s' % (self._FAKE_PORT_MIN,
                                            self._FAKE_PORT_MAX),
                    'RemoteIPAddress': self._FAKE_DEST_IP_PREFIX,
                    'Stateful': True,
                    'IdleSessionTimeout': 0}

        sg_rule = self._create_sg_rule()
        self.assertEqual(expected, sg_rule.to_dict())

    def test_localport(self):
        sg_rule = self._create_sg_rule()
        expected = '%s-%s' % (self._FAKE_PORT_MIN, self._FAKE_PORT_MAX)
        self.assertEqual(expected, sg_rule.LocalPort)

    def test_localport_icmp(self):
        sg_rule = self._create_sg_rule(self._acl('protocol', 'icmp'))
        self.assertEqual('', sg_rule.LocalPort)

    def test_stateful_icmp(self):
        sg_rule = self._create_sg_rule(self._acl('protocol', 'icmp'))
        self.assertFalse(sg_rule.Stateful)

    def test_stateful_ipv6_icmp(self):
        sg_rule = self._create_sg_rule(self._acl('protocol', 'ipv6-icmp'))
        self.assertFalse(sg_rule.Stateful)

    def test_stateful_deny(self):
        sg_rule = self._create_sg_rule(action=self._acl('action', 'deny'))
        self.assertFalse(sg_rule.Stateful)

    def test_stateful_true(self):
        sg_rule = self._create_sg_rule()
        self.assertTrue(sg_rule.Stateful)

    def test_rule_uniqueness(self):
        sg_rule = self._create_sg_rule()
        sg_rule2 = self._create_sg_rule(self._acl('protocol', 'icmp'))

        self.assertEqual([sg_rule], list(set([sg_rule] * 2)))
        self.assertEqual(sorted([sg_rule, sg_rule2]),
                         sorted(list(set([sg_rule, sg_rule2]))))



# Copyright 2015 Cloudbase Solutions Srl
# Copyright 2010-2011 OpenStack Foundation
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

"""Base test case for tests that do not rely on Tempest."""

import contextlib
import logging as std_logging
import os
import os.path
import traceback

import eventlet.timeout
import fixtures
import mock
from os_win import utilsfactory
from oslo_utils import strutils
from oslotest import mock_fixture
import testtools

from networking_hyperv.neutron import config

CONF = config.CONF

LOG_FORMAT = "%(asctime)s %(levelname)8s [%(name)s] %(message)s"

mock_fixture.patch_mock_module()


def bool_from_env(key, strict=False, default=False):
    value = os.environ.get(key)
    return strutils.bool_from_string(value, strict=strict, default=default)


class BaseTestCase(testtools.TestCase):

    _autospec_classes = []

    def setUp(self):
        super(BaseTestCase, self).setUp()
        self.useFixture(mock_fixture.MockAutospecFixture())
        self._patch_autospec_classes()

        self.addCleanup(CONF.reset)
        self.addCleanup(mock.patch.stopall)

        if bool_from_env('OS_DEBUG'):
            _level = std_logging.DEBUG
        else:
            _level = std_logging.INFO
        capture_logs = bool_from_env('OS_LOG_CAPTURE')
        if not capture_logs:
            std_logging.basicConfig(format=LOG_FORMAT, level=_level)
        self.log_fixture = self.useFixture(
            fixtures.FakeLogger(
                format=LOG_FORMAT,
                level=_level,
                nuke_handlers=capture_logs,
            ))

        test_timeout = int(os.environ.get('OS_TEST_TIMEOUT', 0))
        if test_timeout == -1:
            test_timeout = 0
        if test_timeout > 0:
            self.useFixture(fixtures.Timeout(test_timeout, gentle=True))

        if bool_from_env('OS_STDOUT_CAPTURE'):
            stdout = self.useFixture(fixtures.StringStream('stdout')).stream
            self.useFixture(fixtures.MonkeyPatch('sys.stdout', stdout))
        if bool_from_env('OS_STDERR_CAPTURE'):
            stderr = self.useFixture(fixtures.StringStream('stderr')).stream
            self.useFixture(fixtures.MonkeyPatch('sys.stderr', stderr))

        self.addOnException(self.check_for_systemexit)

    def _patch_autospec_classes(self):
        for class_type in self._autospec_classes:
            mocked_class = mock.Mock(autospec=class_type)
            patcher = mock.patch(
                '.'.join([class_type.__module__, class_type.__name__]),
                mocked_class)
            patcher.start()

    def check_for_systemexit(self, exc_info):
        if isinstance(exc_info[1], SystemExit):
            self.fail("A SystemExit was raised during the test. %s"
                      % traceback.format_exception(*exc_info))

    @contextlib.contextmanager
    def assert_max_execution_time(self, max_execution_time=5):
        with eventlet.timeout.Timeout(max_execution_time, False):
            yield
            return
        self.fail('Execution of this test timed out')

    def assertOrderedEqual(self, expected, actual):
        expect_val = self.sort_dict_lists(expected)
        actual_val = self.sort_dict_lists(actual)
        self.assertEqual(expect_val, actual_val)

    def sort_dict_lists(self, dic):
        for key, value in dic.items():
            if isinstance(value, list):
                dic[key] = sorted(value)
            elif isinstance(value, dict):
                dic[key] = self.sort_dict_lists(value)
        return dic

    def assertDictSupersetOf(self, expected_subset, actual_superset):
        """Checks that actual dict contains the expected dict.

        After checking that the arguments are of the right type, this checks
        that each item in expected_subset is in, and matches, what is in
        actual_superset. Separate tests are done, so that detailed info can
        be reported upon failure.
        """
        if not isinstance(expected_subset, dict):
            self.fail("expected_subset (%s) is not an instance of dict" %
                      type(expected_subset))
        if not isinstance(actual_superset, dict):
            self.fail("actual_superset (%s) is not an instance of dict" %
                      type(actual_superset))
        for k, v in expected_subset.items():
            self.assertIn(k, actual_superset)
            self.assertEqual(v, actual_superset[k],
                             "Key %(key)s expected: %(exp)r, actual %(act)r" %
                             {'key': k, 'exp': v, 'act': actual_superset[k]})

    def config(self, **kw):
        """Override some configuration values.

        The keyword arguments are the names of configuration options to
        override and their values.

        If a group argument is supplied, the overrides are applied to
        the specified configuration option group.

        All overrides are automatically cleared at the end of the current
        test by the fixtures cleanup process.
        """
        group = kw.pop('group', None)
        for k, v in kw.items():
            CONF.set_override(k, v, group)


class HyperVBaseTestCase(BaseTestCase):
    def setUp(self):
        super(HyperVBaseTestCase, self).setUp()

        utilsfactory_patcher = mock.patch.object(utilsfactory, '_get_class')
        utilsfactory_patcher.start()
        self.addCleanup(utilsfactory_patcher.stop)



# Copyright 2015 Cloudbase Solutions SRL
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from networking_hyperv.neutron import constants


def get_topic_name(prefix, table, operation):
    """Create a topic name.

    The topic name needs to be synced between the agents.
    The agent will send a fanout message to all of the listening agents
    so that the agents in turn can perform their updates accordingly.

    :param prefix: Common prefix for the agent message queues.
    :param table: The table in question (TUNNEL, LOOKUP).
    :param operation: The operation that invokes notification (UPDATE)
    :returns: The topic name.
    """
    return '%s-%s-%s' % (prefix, table, operation)


class AgentNotifierApi(object):
    """Agent side of the OpenVSwitch rpc API."""

    def __init__(self, topic, client):
        self._client = client
        self.topic_tunnel_update = get_topic_name(topic,
                                                  constants.TUNNEL,
                                                  constants.UPDATE)
        self.topic_lookup_update = get_topic_name(topic,
                                                  constants.LOOKUP,
                                                  constants.UPDATE)

    def _fanout_cast(self, context, topic, method, **info):
        cctxt = self._client.prepare(topic=topic, fanout=True)
        cctxt.cast(context, method, **info)

    def tunnel_update(self, context, tunnel_ip, tunnel_type):
        self._fanout_cast(context,
                          self.topic_tunnel_update,
                          'tunnel_update',
                          tunnel_ip=tunnel_ip,
                          tunnel_type=tunnel_type)

    def lookup_update(self, context, lookup_ip, lookup_details):
        self._fanout_cast(context,
                          self.topic_lookup_update,
                          'lookup_update',
                          lookup_ip=lookup_ip,
                          lookup_details=lookup_details)
